# -*- coding: utf-8 -*-
"""
This script is the cVAE for the publication 
Machine learning of optical properties of materials - predicting spectra from images and images from spectra
by
Helge S. Stein, Dan Guevara, Paul F. Newhouse, Edwin Soedarmadji, John M. Gregoire,
Joint Center for Artificial Photosynthesis
California Institute of Technology
Pasadena
California 91125 (USA)

correspondence should be adressed to :
stein@caltech.edu, gregoire@caltech.edu
"""

import numpy as np
from keras.layers import Input, Dense, Lambda, Reshape, Flatten, Dropout
from keras.layers.merge import concatenate
from keras.models import Model
from keras import backend as K
from keras.datasets import mnist
from keras.utils import to_categorical
from keras.callbacks import EarlyStopping
from scipy.misc import imsave
from keras.utils import HDF5Matrix
import pickle
from keras import metrics
from keras.layers import Conv2D, MaxPooling2D, UpSampling2D, AveragePooling2D
from keras.layers import BatchNormalization, Activation, ZeroPadding2D
from keras.layers.advanced_activations import LeakyReLU, PReLU
import matplotlib.pyplot as plt
from keras.preprocessing.image import ImageDataGenerator
from keras.utils import to_categorical
from sklearn.preprocessing import StandardScaler, MinMaxScaler, QuantileTransformer
from scipy import linalg
from keras import regularizers
from sklearn.model_selection import GridSearchCV
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.utils.validation import check_is_fitted
from sklearn.utils import check_array, as_float_array


def load_data(path):
    #all this does now is to import data preprocessign is done later in the generator
    X_train = HDF5Matrix(path, 'images_train')
    X_test = HDF5Matrix(path, 'images_test')

    y_train = HDF5Matrix(path, 'spectra_train')
    y_test = HDF5Matrix(path, 'spectra_test')


    return X_train, X_test, y_train, y_test

X_train, X_test, y_train, y_test = load_data(r'D:\dataset_publish\test_train_data.h5')

params = {
        'optimizer' : 'adam',
        'm' : 2048,
        'n_x' : X_train.shape[1:], #symmetric
        'n_y' : y_train.shape[1],
        'n_z' : 100,
        'alpha' : 0.001,
        'nepoch' : 1000
        }

## VARIATIONAL ENCODER
def create_vae(pars):

    f1 = pars['f1']
    f2 = pars['f2']
    f3 = pars['f3']
    f4 = pars['f4']
    k = pars['k']
        
    X = Input(shape=pars['n_x'],name='inputimg')
    
    mid = Conv2D(f1, k, activation='relu', padding='same')(X)
    mid = MaxPooling2D((2, 2), padding='same')(mid)
    mid = Conv2D(f2, k, activation='relu', padding='same')(mid)
    mid = MaxPooling2D((2, 2), padding='same')(mid)
    mid = Conv2D(f3, k, activation='relu', padding='same')(mid)
    mid = MaxPooling2D((2, 2), padding='same')(mid)
    mid = Conv2D(f4, k, activation='relu', padding='same')(mid)
    midl = MaxPooling2D((2, 2), padding='same')(mid)
    
    mide = Flatten(name='flattened')(midl)
    mu_a = Dense(pars['n_z'],name='mu',activation='linear')(mide)
    log_sigma_a = Dense(pars['n_z'],name='logsigma',activation='linear')(mide)
    
    def sample_z(args):
        mu, log_sigma = args
        z_rand = mu +  K.random_normal(K.shape(mu),mean=0.0, stddev=1.0) * K.exp(log_sigma/2)
    
        return K.in_train_phase(z_rand,mu)
    
    z = Lambda(sample_z, output_shape = (pars['n_z'], ),
               name='zlayer')([mu_a, log_sigma_a])
    
    encoder = Model(X, z)
    
    decoder_input = Input(shape=(pars['n_z'], ),name='input_decoder')
    ds = midl.shape[1].value*midl.shape[2].value*midl.shape[3].value
    dec_dense = Dense(ds,activation='relu')(decoder_input)
    res_dec_dense = Reshape(target_shape=(midl.shape[1].value,midl.shape[2].value,midl.shape[3].value,))(dec_dense)
    x1 = Conv2D(f1, k, activation='relu', padding='same')(res_dec_dense)
    x2 = UpSampling2D((2, 2))(x1)
    x3 = Conv2D(f2, k, activation='relu', padding='same')(x2)
    x4 = UpSampling2D((2, 2))(x3)
    x5 = Conv2D(f3, k, activation='relu', padding='same')(x4)
    x6 = UpSampling2D((2, 2))(x5)
    x7 = Conv2D(f4, k, activation='relu', padding='same')(x6)
    x8 = UpSampling2D((2, 2))(x7)
    outputs_a = Conv2D(3, k, activation='sigmoid', padding='same')(x8)
    
    decoder = Model(decoder_input,outputs_a)
    
    vae = Model(inputs=encoder.input, outputs=decoder(encoder.output))
    
    def vae_loss(x, out):
        xent_loss =  metrics.binary_crossentropy(x, out)*64*64*3
        kl_loss = - 0.5 * K.mean(1 + log_sigma_a - K.square(mu_a) - K.exp(log_sigma_a),axis=-1)
        return (K.mean(xent_loss) + K.mean(kl_loss))
    
    def recon_loss(x, out):
        return  metrics.binary_crossentropy(x, out)*64*64*3
    
    def KL_loss(inp, outp):
        x_mean, x_log_var = mu_a,log_sigma_a
        kl_loss = - 0.5 * K.mean(1 + x_log_var - K.square(x_mean) - K.exp(x_log_var))
        return kl_loss
    vae.compile(optimizer='adam',
                 loss=vae_loss,
                 metrics = [KL_loss, recon_loss,'mse'])
    
    return encoder,decoder,vae

pars = {'f1':[32,16,8],'f2':[16,8,4],'f3':[4],'f4':[4],'k':[(3,3)],'n_x' : [X_train.shape[1:]], #symmetric
        'n_y' : [y_train.shape[1]],'n_z':[50,100,150]}

import itertools as it

keys = sorted(pars)
combinations = it.product(*(pars[k] for k in keys))
pars_list = [{k:v for k,v in zip(keys,co)} for co in combinations]

res = []
for p in pars_list:
    print(p)
    encoder,decoder,vae = create_vae(p)
    vae.summary()

    vae_hist = vae.fit(X_train,
                         X_train,
                         batch_size=512,
                         epochs=50,
                         shuffle='batch',
    					 validation_data = (X_test,X_test))
    res.append([vae_hist.history,p])
pickle.dump(res,open('E:\grid_save.pck','wb'))  
   
#the 'best' model found:

encoder,decoder,vae = create_vae({'f1':8,'f2':16,'f3':4,'f4':4,'k':(3,3),'n_x' : X_train.shape[1:], #symmetric
        'n_y' : y_train.shape[1],'n_z':100})
  
vae_hist = vae.fit(X_train,
                   X_train,
                   batch_size=512,
                   epochs=100,
                   shuffle='batch',
    				   validation_data = (X_test,X_test))   
    
    
#if we need to load the weights
#encoder.load_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vaeencoder_publ_conv.h5")
#decoder.load_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vaedecoder_publ_conv.h5")

model_json = vae.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vae_publ_conv.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
vae.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vae_publ_conv.h5")
print("Saved CVAE to disk")

model_json = encoder.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vaeencoder_publ_conv.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
encoder.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vaeencoder_publ_conv.h5")
print("Saved ENCODER to disk")

model_json = decoder.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vaedecoder_publ_conv.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
decoder.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\vaedecoder_publ_conv.h5")
print("Saved DECODER to disk")
#normed_spectra = scale(y_test)

#---------------------------------------------------------- prediction

def mse_loss(y_true, y_pred):
    mse =  metrics.mean_squared_error(y_true, y_pred)*10**2
    return mse

def r2_loss(y_true, y_pred):
    yhat =  K.mean(y_true,axis=0) #mean per energy 
    SS_tot = K.sum(K.square(y_true - yhat),axis=0)
    SS_res = K.sum(K.square(y_true - y_pred),axis=0) 
    r2 = K.clip(1 - SS_res/(SS_tot),0,1)
    return K.sum(1-r2)/220

def r2_loss_trad(y_true, y_pred):
    yhat =  K.mean(y_true) #mean per energy 
    SS_tot = K.sum(K.square(y_true - yhat))
    SS_res = K.sum(K.square(y_true - y_pred)) 
    r2 = K.clip(1 - SS_res/(SS_tot),0,1)
    return K.mean(1-r2)

def spec_score(y_true, y_pred):
    yhat =  K.mean(y_true,axis=0) #mean per energy 
    SS_tot = K.sum(K.square(y_true - yhat),axis=0)
    SS_res = K.sum(K.square(y_true - y_pred),axis=0) 
    r2 = K.clip(1 - SS_res/(SS_tot),0,1)
    mse =  metrics.mean_squared_error(y_true, y_pred)*10**2
    return K.sum(1-r2)/220*mse

def spec_score_combi(y_true, y_pred):
    yhat =  K.mean(y_true,axis=0) #mean per energy 
    SS_tot = K.sum(K.square(y_true - yhat),axis=0)
    SS_res = K.sum(K.square(y_true - y_pred),axis=0) 
    r2a = K.clip(1 - SS_res/(SS_tot),0,1)
    
    yhat =  K.mean(y_true)
    SS_tot = K.sum(K.square(y_true - yhat))
    SS_res = K.sum(K.square(y_true - y_pred)) 
    r2b = K.clip(1 - SS_res/(SS_tot),0,1)
    
    
    mse =  metrics.mean_squared_error(y_true, y_pred)*10**2
    return K.mean(1-r2a)*mse*K.mean(1-r2b)


from keras.layers import Conv1D,UpSampling1D
def create_spec(pars):
    e = Input(shape=(100,),name='inputenc')
    d0 = Dense(pars['a'])(e)
    d0 = PReLU()(d0)
    d0 = Dense(55)(d0)
    d0 = PReLU()(d0)
    d0 = Reshape((55,1,))(d0)
    d0 = Conv1D(pars['f1'],pars['k1'],padding='same')(d0)
    d0 = PReLU()(d0)
    d0 = UpSampling1D()(d0)
    d0 = Conv1D(pars['f2'],pars['k2'],padding='same')(d0)
    d0 = PReLU()(d0)
    d0 = UpSampling1D()(d0)
    d0 = Conv1D(1,1,strides=1,activation='linear')(d0)
    out = Flatten()(d0)
    spec = Model(e,out)
    spec.compile(optimizer='adam',
                 loss=spec_score,metrics=[mse_loss,r2_loss])
    spec.summary()
    
    return spec

pars = {'a':[100,50],'f1':[64,32],'f2':[64,32],'k1':[10],'k2':[20]}

import itertools as it

keys = sorted(pars)
combinations = it.product(*(pars[k] for k in keys))
pars_list = [{k:v for k,v in zip(keys,co)} for co in combinations]

enc_test = encoder.predict(X_test)
enc_train = encoder.predict(X_train)

res = []
for j,p in enumerate(pars_list):
    print(p)
    spec = create_spec(p)
    spec.summary()

    spec_hist = spec.fit(enc_train,
                         y_train,
                         batch_size=5000,
                         epochs=100,
                         shuffle='batch',
    						 validation_data = (enc_test,y_test))
    res.append([spec_hist.history,p])

#best is:
optim = {'a': 100, 'f1': 64, 'f2': 32, 'k1': 10, 'k2': 20}
spec = create_spec(optim)
spec.summary()
spec_hist = spec.fit(enc_train,
                         y_train,
                         batch_size=5000,
                         epochs=1000,
                         shuffle='batch',
    						 validation_data = (enc_test,y_test))

model_json = spec.to_json()
with open(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\spec_publ_conv.json", "w") as json_file:
    json_file.write(model_json)
# serialize weights to HDF5
spec.save_weights(r"K:\users\hte\Manuscripts\inpreparation\autoencoder\models\spec_publ_conv.h5")
print("Saved SPEC to disk")

#special losses plot
fig,axs = plt.subplots(1,2,figsize=(10,5))
axs[0].semilogy(spec.history.history['loss'],label='Train loss',c='black',alpha=0.7)
axs[0].semilogy(spec.history.history['val_loss'],label='Val. loss',c='black')
axs[0].set_xlabel('Epoch')
axs[0].set_ylabel('Loss $MSE*(1-R^2)$')
axs[0].legend()
axs[1].plot(spec.history.history['r2_loss'],label='$R^2$ Train loss',c='red',alpha=0.7)
axs[1].plot(spec.history.history['val_r2_loss'],label='$R^2$ Val. loss',c='red')
axs[1].plot(spec.history.history['mse_loss'],label='MSE Train loss',c='orange',alpha=0.7)
axs[1].plot(spec.history.history['val_mse_loss'],label='MSE Val. loss',c='orange')
axs[1].set_xlabel('Epoch')
axs[1].set_ylabel('$R^2$')
axs[1].legend()
plt.savefig(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\loss_spec_pred_conv_cuholdout.svg',dpi=600)

pred_spec = spec.predict(y_test)
err = np.mean((pred_spec-y_test)**2,axis=-1)

r,c = 10,10
fig, axs = plt.subplots(r, c,figsize=(10,10),sharey='all',sharex='all')
idxes = []
for perc in range(1,11):
    lo,up = np.percentile(err,(perc-1)*10),np.percentile(err,perc*10)
    print('p0 {} p1 {}'.format((perc-1)*10,(perc)*10))
    sel = np.where((err>lo) & (err<up))[0]
    idx = np.sort(np.random.choice(sel,10))
    idxes.append(idx)
    for ii,i in enumerate(idx):
        axs[perc-1,ii].plot(pred_spec[i],color='black')
        axs[perc-1,ii].plot(y_test[i],color='green')
        axs[perc-1,ii].axis('off')
plt.subplots_adjust(wspace=0, hspace=0)
pickle.dump(idxes, open(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\indexes.pck','wb'))
plt.savefig(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\spec_pred_percentiles0.svg',dpi=600)

pickle.dump(enc_test, open(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\enc_test.pck','wb'))
pickle.dump(enc_train, open(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\enc_train.pck','wb'))
pickle.dump(y_test, open(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\y_test.pck','wb'))
pickle.dump(np.copy(y_train), open(r'K:\users\hte\Manuscripts\inpreparation\autoencoder\figures\y_train.pck','wb'))

from pyearth import Earth as MARS

def get_bg(X,Y):
    # _s for simple _c_s low high interc _c_z according to Schwarting et al.
    model = MARS()
    model.fit(X,np.array((Y*energ)**2))
    Y_h = model.predict(X)

    diff1 = np.diff(Y_h)/np.diff(X)
    tdiff1 = diff1-np.nanmin(diff1)
    tdiff1 = tdiff1/np.nanmax(tdiff1)
    ID = [i for i in range(1,len(tdiff1)) if np.abs(tdiff1[i]-tdiff1[i-1])>0.02]
    ID.insert(0,0)
    ID.append(np.argmax(X))#this might cause an error
    slopes = np.array([np.nanmean(diff1[ID[i-1]:ID[i]]) for i in range(1,len(ID))])
    a = [Y_h[ID[i]]-slopes[i]*X[ID[i]] for i in range(len(ID)-1)]
    IDM, IDm = np.argmax(slopes), np.argmin(np.abs(slopes))
    #intercept of highest slope and zero as well as highest slope and lowest slope
    bg_s = -a[IDM]/slopes[IDM]
    bg_c_s = (a[IDM]-a[IDm])/(slopes[IDm]-slopes[IDM])

    w = np.array([X[ID[i]]-X[ID[i-1]] for i in range(1,len(ID))])
    d = np.array([Y_h[ID[i]]-Y_h[ID[i-1]] for i in range(1,len(ID))])
    l = np.array(np.sqrt(np.sqrt(w**2+d**2)))

    IDM, IDm = np.argmax(slopes*(w**2)*l*d), np.argmin(np.abs(slopes))
    bg_z = -a[IDM]/slopes[IDM]
    bg_c_z = (a[IDM]-a[IDm])/(slopes[IDm]-slopes[IDM])
    
    return bg_s,bg_c_s,bg_z,bg_c_z

from tqdm import tqdm

bgs_test = np.array([get_bg(energ,Y) for Y in tqdm(y_test)])
bgs_pred = np.array([get_bg(energ,Y) for Y in tqdm(y_pred)])

#filter for misfits
diffs = []
w=2
for i in range(len(bgs_test)):
    if bgs_test[i,w]>1.31:
            if bgs_test[i,w]<3.1:
                    if bgs_pred[i,w]>1.31:
                            if bgs_pred[i,w]<3.1:
                                if np.isfinite(bgs_test[i,w]-bgs_pred[i,w]):
                                    if np.abs(bgs_test[i,w]-bgs_pred[i,w])<3:
                                        diffs.append([bgs_test[i,w],bgs_pred[i,w]])
diffs = np.array(diffs)
from sklearn.metrics import mean_absolute_error,mean_squared_error
print(mean_absolute_error(diffs[:,0].T,diffs[:,1].T))
print(np.sqrt(mean_squared_error(diffs[:,0].T,diffs[:,1].T)))

plt.hist(diffs[:,0]-diffs[:,1],bins=100,range=(-1.5,1.5),color='black')
